/*
 *  $Id: main.cpp,v 1.2 2003/01/05 12:30:02 omalley Exp $
 *  Copyright (c) 2002 Kevin O'Malley. All rights reserved.
 *
 *  Example program from "Programming Mac OS X: A Guide for
 *  UNIX Developers" by Kevin O'Malley, Manning Publishing
 *  2002.
 */
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "BuggyCode.h"
#include "MySockets.h"

const char *PROGRAM_NAME = "BuggyServer";

/*
 Print program messages, which the client formats as printf-style
 calls.
 */
void
Trace(char *fmt, ...)
{
  va_list argp;
  
  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fflush(stdout);
}

/*
 Print the program's usage statement.
 */
void
Usage()
{
  Trace("%s: [port]\n", PROGRAM_NAME);
  exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[])
{
  int sockFD, newSockFD, clientLen;
  struct sockaddr_in	clientAddr, serverAddr;

  if (argc < 2)
    Usage();

  /*
   Grab the port number from the commandline.  We use atoi to
   convert the initial portion of the ASCII string pointed to by
   argv[1] to an integer.
   */
  int port = atoi(argv[1]);

  /*
   * Open a TCP socket.
   */
  if ( (sockFD = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    Trace("%s: can't open stream socket.\n", PROGRAM_NAME);
    exit(EXIT_FAILURE);
  }

  serverAddr.sin_family = AF_INET;
  serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  serverAddr.sin_port = htons(port);

  /*
   We use the setsockopt call, with SO_REUSEADDR, to enable the server
   to bind to an address in a TIME_WAIT state.  See Richard Stevens's
   "W. R. Stevens, UNIX network programming, 2nd ed. Upper Saddle River,
   NJ: Prentice Hall PTR, 1998" for more information.
   */
  int flag = 1;
  int ret = setsockopt(sockFD, SOL_SOCKET, SO_REUSEADDR,
                       (char *)&flag, sizeof(flag));
  if (ret != 0) {
    Trace("%s: setsockopt error", PROGRAM_NAME);
    exit(EXIT_FAILURE);
  }

  /*
   * Bind to local address so that the client can send us messages.
   */
  if (bind(sockFD, (struct sockaddr *) &serverAddr, sizeof(serverAddr)) < 0) {
    Trace("%s: can't bind local address", PROGRAM_NAME);
    exit(EXIT_FAILURE);
  }

  /*
   Set the backlog to 5, which should be enough.  The backlog parameter
   defines the maximum length the queue of pending connections may grow to.
   */
  ret = listen(sockFD, 5);
  if (ret != 0) {
    Trace("%s: listen error", PROGRAM_NAME);
    exit(EXIT_FAILURE);
  }
  
  Trace("%s running on port: %d, pid: %d\n", PROGRAM_NAME,
        port, getpid());

  bool quit = false;
  for (;;) {
    /*
     * Wait for a connection from a client process.
     */
    clientLen = sizeof(clientAddr);
    newSockFD = accept(sockFD, (struct sockaddr *) &clientAddr,
                       &clientLen);
    if (newSockFD < 0) {
      Trace("%s: server: accept error", PROGRAM_NAME);
      exit(EXIT_FAILURE);
    }

    // Setup buffers and read the client message.
    char recvBuf[MAX_STRING_LEN+1];
    char reply[MAX_STRING_LEN+1];
    int nBytes = ReadString(newSockFD,
                            recvBuf,
                            MAX_STRING_LEN);
    if (nBytes == 0) {
      Trace("%s: read 0 bytes\n", PROGRAM_NAME);
      break;
    }

    quit = Process(recvBuf, reply);

    // Write the result back to the client.
    nBytes = WriteString(newSockFD,
                         reply,
                         MAX_STRING_LEN);
    if (nBytes == 0) {
      Trace("%s: wrote 0 bytes\n", PROGRAM_NAME);
      break;
    }

    close(newSockFD);

    if (quit)
      exit(0);
  }
  Trace("done");
}
