/*
 *  $Id: main.cpp,v 1.2 2003/01/05 12:36:46 omalley Exp $
 *  Copyright (c) 2002 Kevin O'Malley. All rights reserved.
 *
 *  Example program from "Programming Mac OS X: A Guide for
 *  UNIX Developers" by Kevin O'Malley, Manning Publishing
 *  2002.
 */
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>

#include "MySockets.h"

const char *PROGRAM_NAME = "SampleServer";
const int DEFAULT_PORT = 4444;

/*
 Print program messages, which the client formats as printf-style
 calls.
 */
void
Trace(char *fmt, ...)
{
  va_list argp;
  
  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fflush(stdout);
}

int
main(int argc, char *argv[])
{
  int			sockfd, newsockfd, clilen;
  struct sockaddr_in	cli_addr, serv_addr;

  int port = DEFAULT_PORT;

  /*
   * Open a TCP socket.
   */
  if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    Trace("%s: can't open stream socket.\n", PROGRAM_NAME);
    exit(EXIT_FAILURE);
  }

  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(port);

  /*
   We use the setsockopt call, with SO_REUSEADDR, to enable the server
   to bind to an address in a TIME_WAIT state.  See Richard Stevens's
   "W. R. Stevens, UNIX network programming, 2nd ed. Upper Saddle River,
   NJ: Prentice Hall PTR, 1998" for more information.
   */
  int flag = 1;
  int ret = setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
                       (char *)&flag, sizeof(flag));
  if (ret != 0) {
    Trace("%s: setsockopt error", PROGRAM_NAME);
    exit(EXIT_FAILURE);
  }

  /*
   * Bind to local address so that the client can send us messages.
   */
  if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
    Trace("%s: can't bind local address", PROGRAM_NAME);
    exit(EXIT_FAILURE);
  }

  /*
   Set the backlog to 5, which should be enough.  The backlog parameter
   defines the maximum length the queue of pending connections may grow to.
   */
  listen(sockfd, 5);

  Trace("%s running on port: %d, pid: %d\n", PROGRAM_NAME,
        port, getpid());

  int cnt = 0;
  for (;;) {
    cnt++;
    /*
     * Wait for a connection from a client process.
     * This is an example of a iterative server.
     */
    clilen = sizeof(cli_addr);
    newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr,
                       &clilen);
    if (newsockfd < 0) {
      Trace("%s: server: accept error", PROGRAM_NAME);
      exit(EXIT_FAILURE);
    }

    int     flag = 1;
    int ret = setsockopt(newsockfd, SOL_SOCKET, SO_REUSEADDR,
                         (char *)&flag, sizeof(int));
    if (ret != 0) {
      Trace("%s: setsockopt error", PROGRAM_NAME);
      exit(EXIT_FAILURE);
    }

    // Setup buffers and read the client message.
    char recvBuf[MAX_STRING_LEN+1];
    int nBytes = ReadString(newsockfd,
                            recvBuf,
                            MAX_STRING_LEN);
    if (nBytes <= 0) {
      Trace("%s: read %d bytes\n", nBytes, PROGRAM_NAME);
      break;
    }

    // Write the result back to the client.
    Trace("%s", recvBuf);
    nBytes = WriteString(newsockfd,
                         recvBuf,
                         MAX_STRING_LEN);
    if (nBytes == 0) {
      Trace("%s: wrote 0 bytes\n", PROGRAM_NAME);
      break;
    }

    close(newsockfd);		/* parent process */
    if (cnt == 100)
      break;	
  }
  Trace("done");
}
