/*
 *  Copyright (c) 2002 Kevin O'Malley. All rights reserved.
 *
 *  Example program from "Programming Mac OS X: A Guide for
 *  UNIX Developers" by Kevin O'Malley, Manning Publishing
 *  2002.
 */
static char const rcsid[] = "$Id: main.cpp,v 1.2 2003/01/05 12:41:24 omalley Exp $";

#include <pthread.h>
#include <unistd.h>
#include <cstdlib>
#include <fcntl.h>
#include <signal.h>

#include <cerrno>
#include <ctype.h>
#include <cstdio>
#include <ctime>
#include <stdarg.h>

#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <strings.h>

using namespace std;

#include <string>
#include <vector>
#include <list>
#include <regex.h>

#include <netdb.h>

#include "MySockets.h"

const unsigned int NUM_THREADS = 5;
const unsigned int PORT = 4444;

// The thread control block.
typedef struct {

  // General mutex.
  pthread_mutex_t mutex;
  // Used to lock access to socket connections.
  pthread_mutex_t socketMutex;
  // Used to lock access to connection pool.
  pthread_mutex_t connectionPoolMutex;

  // Thread condition variables.
  pthread_cond_t less;
  pthread_cond_t connectionPoolLess;

  list<int> connectPool;

  unsigned int    threadCnt;
  unsigned int    eventCnt;
}ThreadControlBlock;

ThreadControlBlock *mTCB;
vector<int>        mThreads;

int ProcessMessage(int savedCP, long &remoteUserID,
                   bool &quit, char *recvBuf);

/*
 Init the thread data structures.
 */
bool
TCBInit()
{
  mTCB = (ThreadControlBlock *) new ThreadControlBlock;
  if (mTCB == NULL)
    return false;

  // Init mutex's
  int ret = pthread_mutex_init(&mTCB->mutex, NULL);
  if (ret != 0) {
    delete mTCB;
    return false;
  }
  ret = pthread_mutex_init(&mTCB->socketMutex, NULL);
  if (ret != 0) {
    delete mTCB;
    return false;
  }
  ret = pthread_mutex_init(&mTCB->connectionPoolMutex, NULL);
  if (ret != 0) {
    delete mTCB;
    return false;
  }

  // Init condition variables
  ret = pthread_cond_init(&mTCB->connectionPoolLess, NULL);
  if (ret != 0) {
    delete mTCB;
    return false;
  }

  ret = pthread_cond_init(&mTCB->less, NULL);
  if (ret != 0) {
    delete mTCB;
    return false;
  }

  mTCB->threadCnt = 0;
  mTCB->eventCnt = 0;
  return true;
}

/*
 De-allocate the thread data structures.
 */
void
TCBDelete()
{
  int ret = pthread_mutex_destroy(&mTCB->mutex);
  if (ret != 0)
    printf("TCBDelete: pthread_mutex_destroy:mutex ! return 0: %d\n", ret);
  
  ret = pthread_mutex_destroy(&mTCB->socketMutex);
  if (ret != 0)
    printf("TCBDelete: pthread_mutex_destroy:socketMutex ! return 0: %d\n", ret);

  ret = pthread_mutex_destroy(&mTCB->connectionPoolMutex);
  if (ret != 0)
    printf("TCBDelete: pthread_mutex_destroy:connectionPoolMutex ! return 0: %d\n", ret);

  ret = pthread_cond_destroy(&mTCB->less);
  if (ret != 0)
    printf("TCBDelete: pthread_cond_destroy:less ! return 0: %d\n", ret);

  ret = pthread_cond_destroy(&mTCB->connectionPoolLess);
  if (ret != 0)
    printf("TCBDelete: pthread_cond_destroy:connectionPoolLess ! return 0: %d\n", ret);

  delete mTCB;
}

/*
 Process the messages.
 */
extern "C"
void *ProcessRequests(void *arg)
{
  arg = arg;

  char recvBuf[MAX_STRING_LEN + 1] = "";
  char recvOrgBuf[MAX_STRING_LEN + 1] = "";

  for(;;) {
    pthread_mutex_lock(&mTCB->connectionPoolMutex);

    while (mTCB->connectPool.size() == 0)
      pthread_cond_wait(&mTCB->connectionPoolLess,
                        &mTCB->connectionPoolMutex);

    int savedCP = mTCB->connectPool.front();
    mTCB->connectPool.pop_front();
    mTCB->threadCnt++;

    /*
     	Commenting out the following statement will
     	cause the program to block on new client
     	connections.
     */
    pthread_mutex_unlock(&mTCB->connectionPoolMutex);

    // I am calling this to make sure the ports get reused.
    int flag = 1;
    setsockopt(savedCP, SOL_SOCKET, SO_REUSEADDR,
               (char *)&flag, sizeof(flag));

    int  nRead = 0;
    for(;;) {
      nRead = 0;
      strcpy(recvOrgBuf, "");
      nRead = ReadString(savedCP, recvOrgBuf,
                         MAX_STRING_LEN);
      if (nRead == 0)
        break;

      int i=0, j=0;
      bool messageLoopDone = false;
      while (!messageLoopDone) {
        j = 0;
        for(i=i; i<nRead; i++) {
          recvBuf[j++] = recvOrgBuf[i];
          if (recvOrgBuf[i] == 0)
            break;
        }

        if (strlen(recvBuf) != 0) {
          printf("ProcessMessage: %s\n", recvBuf);
          fflush(stdout);
          
          int nBytes = WriteString(savedCP, recvBuf,
                                   (int)(strlen(recvBuf)));
          if (nBytes == 0) {
            printf("read zero bytes\n");
            fflush(stdout);
         }
          break;
        }
        // Get's us past the NULL from recvOrgBuf.
        i++;
        if (i >= nRead)
          messageLoopDone = true;
      } // while
    } // inner for

    pthread_mutex_lock(&mTCB->mutex);
    mTCB->threadCnt--;

    /*
     	Commenting out the following statement will
     	cause the program to block on new client
     	connections.
     */
    pthread_mutex_unlock(&mTCB->mutex);

    close(savedCP);
  } // other for
}

/*
 Process a message.
 */
int
ProcessMessage(int savedCP, long &remoteUserID,
               bool &quit, char *recvBuf)
{
  printf("ProcessMessage: %s\n", recvBuf);
  fflush(stdout);
  
  char buf[MAX_STRING_LEN + 1];
  int nBytes = WriteString(savedCP, buf,
                           (int)(strlen(buf)));
  if (nBytes == 0) {
    printf("read zero bytes\n");
    fflush(stdout);
  }
  if (strcmp(recvBuf, "quit") == 0) {
    return 1;
  }
  return 0;
}

/*
 Handle signals.
 */
void
SigHandler(int signum)
{
  if (signum == SIGPIPE) {
  }
}

int
main(int argc, char *argv[])
{
  int sockFD, newSockFD, clientLen;
  struct sockaddr_in clientAddr, serverAddr;

  /*
   Setup the signal handler to ignore the SIGPIPE
   signal, #13.
   */
  struct sigaction sig;
  sig.sa_handler = SigHandler;
  sigemptyset (&sig.sa_mask);

  if (sigaction(SIGPIPE, &sig, NULL) < 0) {
    printf("api: error installing signal handler\n");
    exit(EXIT_FAILURE);
  }

  /*
   * Open a TCP socket.
   */
  if ( (sockFD = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    printf("main: error opening socket\n");
    exit(EXIT_FAILURE);
  }

  /*
   We use the setsockopt call, with SO_REUSEADDR, to enable the server
   to bind to an address in a TIME_WAIT state.  See Richard Stevens's
   "W. R. Stevens, UNIX network programming, 2nd ed. Upper Saddle River,
   NJ: Prentice Hall PTR, 1998" for more information.
   */
  int flag = 1;
  int ret = setsockopt(sockFD, SOL_SOCKET, SO_REUSEADDR,
                       (char *)&flag, sizeof(flag));
  if (ret != 0) {
    printf("main: can't set socket to reuse\n");
    exit(EXIT_FAILURE);
  }

  bzero((char *) &serverAddr, sizeof(serverAddr));
  serverAddr.sin_family = AF_INET;
  serverAddr.sin_port = htons(PORT);
  serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);

  /*
   * Bind to local address so that the client can send us messages.
   */
  if (bind(sockFD, (struct sockaddr *)
           &serverAddr, sizeof(serverAddr)) < 0) {
    printf("main: can't bind to port, maybe another process is running on the port\n");
    exit(EXIT_FAILURE);
  }

  /*
   Set the backlog to 5, which should be enough.  The backlog parameter
   defines the maximum length the queue of pending connections may grow to.
   */
  if (listen(sockFD,  5) != 0) {
    printf("main: error listen\n");
    exit(EXIT_FAILURE);
  }
  
  if (!TCBInit()) {
    printf("main: error initializing thread control block\n");
    exit(EXIT_FAILURE);
  }

  // Create the connection thread pool.
  pthread_t thr_t;
  pthread_attr_t attr;
  ret = pthread_attr_init(&attr);
  if (ret != 0) {
    printf("main: pthread_attr_init error: %d\n", ret);
    exit(EXIT_FAILURE);
  }
  
  ret = pthread_attr_setdetachstate(&attr,
                                    PTHREAD_CREATE_DETACHED);
  if (ret != 0) {
    printf("main: pthread_attr_setdetachstate error: %d\n", ret);
    exit(EXIT_FAILURE);
  }
  
  for(unsigned int i=0; i<NUM_THREADS; i++) {
    ret = pthread_create(&thr_t,
                         &attr,
                         ProcessRequests,
                         (void*)NULL);
    if (ret != 0) {
      printf("main: pthread_create error: %d\n", ret);
      exit(EXIT_FAILURE);
    }
    mThreads.push_back((int)thr_t);
  }
  pthread_attr_destroy(&attr);

  printf("main: running server on port: %d, thread pool size: %d\n",
         PORT, NUM_THREADS);
  fflush(stdout);
 
  /*
   Save this so we can close it on an undoc_apiExit
   messsage.
   */
  for (;;) {
    /*
     Wait for a connection from a client process.
     This is an example of a concurrent server.
     */
    for (;;) {
      clientLen = sizeof(clientAddr);
      newSockFD = accept(sockFD,
                         (struct sockaddr *) &clientAddr,
                         &clientLen);
      if (newSockFD < 0) {
        if (errno == EINVAL) {
          continue;
        }
      }
      else
        break;
    }

    if (newSockFD >= 0) {
      pthread_mutex_lock(&mTCB->connectionPoolMutex);
    
      mTCB->connectPool.push_back(newSockFD);
      pthread_cond_signal(&mTCB->connectionPoolLess);

      pthread_mutex_unlock(&mTCB->connectionPoolMutex);
    }
  } /* for */
  
  TCBDelete();
}

